"use strict";

const t = require("@babel/types");
const template = require("@babel/template");
const LIMIT_LINE = 0;
const ErrorCapture = 'ErrorCapture';

const visitor = {
  ['FunctionExpression|FunctionDeclaration|ArrowFunctionExpression|ClassMethod|ObjectMethod'](path, state) {
    // Function 节点
    let node = path.node,
      params = node.params,
      blockStatement = node.body,
      // 函数function内部代码，将函数内部代码块放入 try 节点
      isGenerator = node.generator,
      isAsync = node.async;
    
    // =================================== 边界情况 return 处理 ============================
    // 1、如果有try catch包裹了，则不需要
    // 2、防止 circle loops，因为节点被替换后会再次进入新节点的遍历
    // 3、需要 try catch 的只能是语句，像 () => 0 这种的 body，是不需要的
    // 4、如果函数内容小于等于 LIMIT_LINE 行不去 try/catch，当然这个函数可以暴露出来给用户设置
    if (
      blockStatement.body && t.isTryStatement(blockStatement.body[0]) ||
      !t.isBlockStatement(blockStatement) && !t.isExpressionStatement(blockStatement) ||
      blockStatement.body && blockStatement.body.length <= LIMIT_LINE ||
      node.id && node.id.name === ErrorCapture
    ) {
      return;
    }
    
    // 创建 catch 节点中的代码
    let catchStatement = template.statement(ErrorCapture + "(error)")();
    let catchClause = t.catchClause(
      t.identifier('error'),
      t.blockStatement([catchStatement])
    );
    
    // 创建 try/catch 的 ast
    let tryStatement = t.tryStatement(blockStatement, catchClause);
    // 创建新节点
    let func = null;
    // 判断：类方法、对象方法、函数申明、函数表达式
    if (t.isClassMethod(node)) {
      // 用 t.classMethod 生成 ast
      func = t.classMethod(node.kind, node.key, params, t.BlockStatement([tryStatement]), node.computed, node.static); // t.BlockStatement([tryStatement]) 新的 ast 节点
    } else if (t.isObjectMethod(node)) {
      func = t.objectMethod(node.kind, node.key, params, t.BlockStatement([tryStatement]), node.computed);
    } else if (t.isFunctionDeclaration(node)) {
      func = t.functionDeclaration(node.id, params, t.BlockStatement([tryStatement]), isGenerator, isAsync);
    } else if (t.isFunctionExpression(node)) {
      func = t.functionExpression(node.id, params, t.BlockStatement([tryStatement]), isGenerator, isAsync);
    } else if (t.isArrowFunctionExpression(node)) {
      func = t.arrowFunctionExpression(params, t.BlockStatement([tryStatement]), isAsync);
    }
    
    // 替换原节点
    path.replaceWith(func);
  }
}

module.exports = visitor;